;;; Package con funciones para implementacion de los 
;;; algoritmo A y A*
;;;
;;; A utilizar con el package "q_search.lsp" (a generic
;;; framework for search) para implementar dichos algoritmos.
;;;
;;; Asignatura: Introduccin a la Inteligencia Artificial
;;;             ETSII - ULL
;;;
;;; Profesor: Patricio Garca Bez (patricio@etsii.ull.es)
;;;
;;; Fecha: 17/5/2002
;;;
;;; Comentarios:
;;;           
;;;           Para la correcta utilizacin han de almacenarse
;;;           en las siguientes variables los nombres de las
;;;           funciones especficas a cada problema:
;;;           
;;;           *EVAL-FN* : 
;;;                       (Definida en q_search.lsp) Funcin que
;;;                       evalua el coste estimado de la solucin
;;;                       mas cercana (f') que pasa por un estado 
;;;                       dado (g + h'). Si se desea implementar 
;;;                       el algoritmo A* esta funcin ha de ser 
;;;                       subestimada.
;;;                       Conviene que sea la funcin que acceda
;;;                       a un campo ya precalculado del estado.
;;;           
;;;           *COSTE-ACUM-FN* :
;;;                       Funcin que calcula el coste acumulado (g)
;;;                       para llegar a un estado dado desde el
;;;                       estado inicial.
;;;           
;;;           *IGUAL-EST-FN* :
;;;                       Predicado que compara si dos estados son
;;;                       iguales o no.
;;;
;;;           Ver el correspondiente help para cada una de ellas.
;;;           
;;;           Las variables anteriores son utilizadas por las dos
;;;           funciones genricas que aqu se definen:  
;;;           PRIORITY-A-MERGE-FN y PRIORITY-A-DEQUEUE-FN
;;;           y que servirn respectivamente como valores de los 
;;;           parmetros :merge-fn y :dequeue-fn para la funcin
;;;           GENERIC-SEARCH definida en el package "q_search.lsp".
;;;           
;;;           Tener en cuenta que en las anteriores funciones no se
;;;           ha impelmentado el tratamiento de estados visitados,
;;;           a parte de aquellos que se encuentran repetidos en 
;;;           la lista frente, siendo responsabilidad del usuario
;;;           su implementacin si as lo cree necesario.
;;;
;;; Modificiaciones:
;;;           Mayo 2004 - Adaptarse a q_search en forma de paquete
;;;                     - No utilizar ms pares punteados en 
;;;                       *search-queue*
;;;           Mayo 2006 - Ampliada documentacin. Definicin como
;;;                       package.
;;;


;;; Definicin del package
;;; 

(defpackage "A_STAR"
  (:use "LISP" "Q_SEARCH")
  (:documentation "Funciones para implementacion de los algoritmo A y A*.")
  (:export "*VERSION-A_STAR*" "*COSTE-ACUM-FN*" "*IGUAL-EST-FN*" "CRIBA-IGUALES" "FUSION" "PRIORITY-A-MERGE-FN" "PRIORITY-A-DEQUEUE-FN"))

(in-package "A_STAR")


;;; Variable versin
;;;

(defconstant *version-a_star* "may-2006" 
  "Versin del programa a_star.")


;;; Variables globales a definir
;;;

(defvar *coste-acum-fn* nil
  "Guarda la funcin que devuelve el coste acumulado (g) de un estado.")

(defvar *igual-est-fn* nil
  "Guarda la funcin que compara dos estados y dice si son iguales.")


;;; Funciones necesarias
;;;

(defun igual-y-peor-o-igual (est1 est2)
  "Indica si el estado EST1 es igual al estado EST2, pero con coste acumulado peor o igual."
  (and (funcall *igual-est-fn* est1 est2)
       (>= (funcall *coste-acum-fn* est1)
	   (funcall *coste-acum-fn* est2))))

(defun igual-y-peor (est1 est2)
  "Indica si el estado EST1 es igual al estado EST2, pero con coste acumulado peor."
  (and (funcall *igual-est-fn* est1 est2)
       (> (funcall *coste-acum-fn* est1)
	  (funcall *coste-acum-fn* est2))))

(defun criba-iguales (new-states)
  "Busca si hay elementos de la lista de hijos (NEW-STATES) en la lista de frente mediante la funcin almacenada en *igual-est-fn* y si es as criba los que tengan coste acumulado (funcin almacenada en *COSTE-ACUM-FN*) mayor de la lista de hijos o de la lista de frente (*SEARCH-QUEUE*)."
  ; Elimina los elementos de new-states que sean iguales a alguno
  ; de la lista frente pero con coste acumulado mayor o igual
  (dolist (el *search-queue*)
    (setf new-states 
	  (delete-if #'(lambda (est) 
			 (igual-y-peor-o-igual est
					       el))
		     new-states 
		     :count 1)))
  ; Elimina los elementos de *search-queue* que sean iguales a alguno
  ; de la new-states pero con coste acumulado mayor
  (dolist (el new-states new-states)
    (setf *search-queue* 
	  (delete-if #'(lambda (est) 
			 (igual-y-peor est el))
		     *search-queue*
		     :count 1))))

(defun fusion (list1 list2)
  "Realiza la fusin ordenada segun f' (*EVAL-FN*) de la lista de hijos (NEW-STATES) en lista frente (*SEARCH-QUEUE*)."
  (cond
   ((endp list1) list2)
   ((endp list2) list1)
   ((<= (funcall *eval-fn* (first list1))
	(funcall *eval-fn* (first list2)))
    (setf (rest list1) (fusion (rest list1) list2))
    list1)
   (t (setf (rest list2) (fusion list1 (rest list2)))
      list2)))

(defun priority-a-merge-fn (new-states)
  "Fusiona la lista de hijos (NEW-STATES) a la lista de frente (*SEARCH-QUEUE*), mantniendola ordenada segun f' (*EVAL-FN*) y eliminando los peores estados de los duplicados."
  ; Chequea y criba si los estados hijos ya estan
  ; en la lista de frente de una u otra lista
  (setf new-states (criba-iguales new-states))
  ; Ordena los estados hijos en funcin de f'
  (setf new-states (sort new-states #'< :key *eval-fn*))
  ; Realiza fusin ordenada de hijos en lista frente
  (fusion new-states *search-queue*))

(defun priority-a-dequeue-fn ()
  "Funcin que extrae el primer elemento de la lista de frente (*SEARCH-QUEUE*) creada con PRIORITY-A-MERGE-FN devolviendo un estado."
  (pop *search-queue*))


;;; *** A* Search ***
;;; Branch and Bound with Underestimates and Dynamic Programming.
#|
(generic-search initial-states goal-p children
		:merge-fn #'priority-a-merge-fn
		:dequeue-fn #'priority-a-dequeue-fn)
|#
